﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Globalization;
using System.IO;
using System.Linq;
using System.Runtime.Serialization.Json;
using System.Xml;
using System.Xml.Linq;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.Testing;

namespace OpcLabs.UnitTestLib.Testers
{
    public class DataContractJsonSerializationTester : Tester
    {
        public DataContractJsonSerializationTester()
        {
            TestCategories.Add("DataContractJsonSerialization");
            TestCategories.Add("Serialization");
        }

        public DataContractJsonSerializationTester([NotNull] Action<object, object> assertion)
            : this()
        {
            Assertion = assertion;
        }

        [CanBeNull]
        public Action<object, object> Assertion { get; }

        protected override void Initialize(Type typeUnderTest)
        {
            base.Initialize(typeUnderTest);

            // ReSharper disable once LoopCanBePartlyConvertedToQuery
            foreach (KeyValuePair<string, CreatorEntry<InstancesAttribute>> instanceCreator in InstanceCreators)
            {
                string instanceName = instanceCreator.Key;
                Debug.Assert(instanceName != null);

                AddTestMethod(CreateNoBackingFieldTestMethod(instanceName));
                AddTestMethod(CreateNoEvaluatingInvariantTestMethod(instanceName));
                AddTestMethod(CreateNoUnderscoredElementsTestMethod(instanceName));
                AddTestMethod(CreateRoundtripTestMethod(instanceName));
            }
        }

        [NotNull]
        private TestMethod CreateNoBackingFieldTestMethod([NotNull] string instanceName)
        {
            return new DelegatedTestMethod(
                string.Format(CultureInfo.InvariantCulture, "NoBackingField_{0}", instanceName),
                () =>
                {
                    object instance = CreateInstance(instanceName);
                    using (var memoryStream = new MemoryStream())
                    {
                        var serializer = new DataContractJsonSerializer(instance.GetType()/*, knownTypes*/);
                        serializer.WriteObject(memoryStream, instance);

                        memoryStream.Position = 0;
                        using (var streamReader = new StreamReader(memoryStream))
                        {
                            string readToEnd = streamReader.ReadToEnd();
                            AssertUtilities.WriteLine(readToEnd);

                            const string backingFieldSignal = "_x003E_k__BackingField";
                            Assert.IsFalse(readToEnd.Contains(backingFieldSignal), $"Must not contain: \"{backingFieldSignal}\"");
                        }
                    }
                });
        }

        [NotNull]
        private TestMethod CreateNoEvaluatingInvariantTestMethod([NotNull] string instanceName)
        {
            return new DelegatedTestMethod(
                string.Format(CultureInfo.InvariantCulture, "NoEvaluatingInvariant_{0}", instanceName),
                () =>
                {
                    object instance = CreateInstance(instanceName);
                    using (var memoryStream = new MemoryStream())
                    {
                        var serializer = new DataContractJsonSerializer(instance.GetType()/*, knownTypes*/);
                        serializer.WriteObject(memoryStream, instance);

                        memoryStream.Position = 0;
                        using (var streamReader = new StreamReader(memoryStream))
                        {
                            string readToEnd = streamReader.ReadToEnd();
                            AssertUtilities.WriteLine(readToEnd);

                            const string evaluatingInvariantSignal = "_x0024_evaluatingInvariant_x0024_";
                            Assert.IsFalse(readToEnd.Contains(evaluatingInvariantSignal), $"Must not contain: \"{evaluatingInvariantSignal}\"");
                        }
                    }
                });
        }

        [NotNull]
        private TestMethod CreateNoUnderscoredElementsTestMethod([NotNull] string instanceName)
        {
            return new DelegatedTestMethod(
                string.Format(CultureInfo.InvariantCulture, "NoUnderscoredElements_{0}", instanceName),
                () =>
                {
                    object instance = CreateInstance(instanceName);
                    using (var memoryStream = new MemoryStream())
                    {
                        var serializer = new DataContractJsonSerializer(instance.GetType()/*, knownTypes*/);
                        serializer.WriteObject(memoryStream, instance);

                        using (var streamReader = new StreamReader(memoryStream))
                        {
                            memoryStream.Position = 0;
                            string readToEnd = streamReader.ReadToEnd();
                            AssertUtilities.WriteLine(readToEnd);

                            memoryStream.Position = 0;
                            XDocument doc = XDocument.Load(streamReader);
                            IEnumerable<XElement> descendants = doc.Descendants();
                            IEnumerable<XElement> elements = descendants.Where(element => 
                                // ReSharper disable once PossibleNullReferenceException
                                element.NodeType == XmlNodeType.Element);

                            XElement offendingElement = elements.FirstOrDefault(element => 
                                // ReSharper disable once PossibleNullReferenceException
                                element.Name.LocalName.StartsWith("_"));

                            Assert.IsNull(offendingElement);
                        }
                    }
                });
        }

        [NotNull]
        private TestMethod CreateRoundtripTestMethod([NotNull] string instanceName)
        {
            return new DelegatedTestMethod(
                string.Format(CultureInfo.InvariantCulture, "Roundtrip_{0}", instanceName),
                () =>
                    {
                        object instance = CreateInstance(instanceName);
                        if (Assertion != null)
                            Tests.DataContractJsonSerializationTest(instance, Assertion);
                        else
                            Tests.DataContractJsonSerializationTest(instance);
                    });
        }
    }
}
