﻿using System.Collections.Generic;
using System.Linq;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.Internal;
using OpcLabs.BaseLib.Reactive;
using OpcLabs.UnitTestLib;

namespace OpcLabs.BaseLibTest.Reactive
{
    [TestClass]
    public class MessageOrderingObserverTest
    {
        [NotNull]
        static private readonly CircularCounterType TwoDigitCircularCounterType =
            new CircularCounterType(0, 99);

        static private void Test(
            int reorderingAreaSize,
            [NotNull] IEnumerable<long> input) =>
            // ReSharper disable PossibleMultipleEnumeration
            Test(reorderingAreaSize, input, input);
        // ReSharper restore PossibleMultipleEnumeration

        static private void Test(
            int reorderingAreaSize,
            [NotNull] IEnumerable<long> input,
            [NotNull] IEnumerable<long> outputBeforeCompleted) =>
            Test(reorderingAreaSize, input, outputBeforeCompleted, Enumerable.Empty<long>());

        static private void Test(
            int reorderingAreaSize,
            [NotNull] IEnumerable<long> input,
            [NotNull] IEnumerable<long> outputBeforeCompleted,
            [NotNull] IEnumerable<long> outputAfterCompleted)
        {
            bool isCompleted = false;

            using (IEnumerator<long> outputBeforeCompletedEnumerator = outputBeforeCompleted.GetEnumerator())
            using (IEnumerator<long> outputAfterCompletedEnumerator = outputAfterCompleted.GetEnumerator())
            {
                var target = new MessageOrderingObserver<long>(
                    TwoDigitCircularCounterType,
                    number => number,
                    reorderingAreaSize,
                    new ActionObserver<long>(value =>
                    {
                        AssertUtilities.WriteLine($"        -> {value}");
                        // ReSharper disable once AccessToModifiedClosure
                        IEnumerator<long> enumerator = isCompleted 
                            // ReSharper disable AccessToDisposedClosure
                            ? outputAfterCompletedEnumerator : outputBeforeCompletedEnumerator;
                        // ReSharper restore AccessToDisposedClosure
                        Assert.IsTrue(enumerator.MoveNext());
                        Assert.AreEqual(value, enumerator.Current);
                    }),
                    droppedMessageObserver:NullObserver<long>.Instance);

                foreach (long value in input)
                {
                    AssertUtilities.WriteLine($"{value} ->");
                    target.OnNext(value);
                }
                Assert.IsFalse(outputBeforeCompletedEnumerator.MoveNext());

                isCompleted = true;
                AssertUtilities.WriteLine("***");
                target.OnCompleted();
                Assert.IsFalse(outputAfterCompletedEnumerator.MoveNext());
            }
        }

        [TestMethod]
        public void Test_RAS0_0() => Test(0, new long[0]);

        [TestMethod]
        public void Test_RAS0_1() => Test(0, new long[] { 1 });

        [TestMethod]
        public void Test_RAS0_2() => Test(0, new long[] { 1, 1 }, new long[] {1});

        [TestMethod]
        public void Test_RAS0_3() => Test(0, new long[] { 1, 2, 3, 4, 5 });

        [TestMethod]
        public void Test_RAS0_4() => Test(0, new long[] { 1, 2, 4, 5 });

        [TestMethod]
        public void Test_RAS3_0() => Test(3, new long[0]);

        [TestMethod]
        public void Test_RAS3_1() => Test(3, new long[] { 1 });

        [TestMethod]
        public void Test_RAS3_2() => Test(3, new long[] { 1, 1 }, new long[] { 1 });

        [TestMethod]
        public void Test_RAS3_3() => Test(3, new long[] { 1, 2, 3, 4, 5 });

        [TestMethod]
        public void Test_RAS3_4() => Test(3, new long[] { 1, 2, 4, 5 }, new long[] {1, 2}, new long [] {4, 5});

        [TestMethod]
        public void Test_RAS3_5() => Test(3, new long[] { 1, 2, 5, 4 }, new long[] { 1, 2 }, new long[] { 4, 5 });

        [TestMethod]
        public void Test_RAS3_6() => Test(3, new long[] { 1, 2, 5, 4, 3 }, new long[] { 1, 2, 3, 4, 5 });

        [TestMethod]
        public void Test_RAS3_7() => Test(3, new long[] { 1, 2, 4, 3, 5 }, new long[] { 1, 2, 3, 4, 5 });

        [TestMethod]
        public void Test_RAS3_8() => Test(3, new long[] { 1, 2, 1, 2 }, new long[] { 1, 2 });

        [TestMethod]
        public void Test_RAS3_9() => Test(3, new long[] { 99, 0 });

        [TestMethod]
        public void Test_RAS3_10() => Test(3, new long[] { 99, 0, 98, 99 }, new long[] { 99, 0 });

        [TestMethod]
        public void Test_RAS3_11() => Test(3, new long[] { 99, 0, 98, 99, 0, 1 }, new long[] { 99, 0, 1 });

        [TestMethod]
        public void Test_RAS3_12() => Test(3, new long[] { 99, 39, 79, 19, 59}, new long[] { 99, 39 }, new long[] {79, 19, 59});

        [TestMethod]
        public void Test_RAS3_13() => Test(3, new long[] { 99, 39, 79, 19, 59, 99 }, new long[] { 99, 39, 79 }, new long[]{19, 59, 99});
    }
}
